# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OpenCensus Collector Metrics Exporter."""

import logging
from typing import Dict, Sequence

import grpc
from google.protobuf.timestamp_pb2 import Timestamp
from opencensus.proto.agent.metrics.v1 import (
    metrics_service_pb2,
    metrics_service_pb2_grpc,
)
from opencensus.proto.metrics.v1 import metrics_pb2
from opencensus.proto.resource.v1 import resource_pb2

import opentelemetry.exporter.opencensus.util as utils
from opentelemetry.sdk.metrics import Counter, Metric
from opentelemetry.sdk.metrics.export import (
    ExportRecord,
    MetricsExporter,
    MetricsExportResult,
)

DEFAULT_ENDPOINT = "localhost:55678"

# In priority order. See collector impl https://bit.ly/2DvJW6y
_OT_LABEL_PRESENCE_TO_RESOURCE_TYPE = (
    ("container.name", "container"),
    ("k8s.pod.name", "k8s"),
    ("host.name", "host"),
    ("cloud.provider", "cloud"),
)

logger = logging.getLogger(__name__)


# pylint: disable=no-member
class OpenCensusMetricsExporter(MetricsExporter):
    """OpenCensus metrics exporter.

    Args:
        endpoint: OpenCensus Collector receiver endpoint.
        service_name: Name of Collector service.
        host_name: Host name.
        client: MetricsService client stub.
    """

    def __init__(
        self,
        endpoint: str = DEFAULT_ENDPOINT,
        service_name: str = None,
        host_name: str = None,
        client: metrics_service_pb2_grpc.MetricsServiceStub = None,
    ):
        self.endpoint = endpoint
        if client is None:
            channel = grpc.insecure_channel(self.endpoint)
            self.client = metrics_service_pb2_grpc.MetricsServiceStub(
                channel=channel
            )
        else:
            self.client = client

        self.node = utils.get_node(service_name, host_name)
        self.exporter_start_timestamp = Timestamp()
        self.exporter_start_timestamp.GetCurrentTime()

    def export(
        self, export_records: Sequence[ExportRecord]
    ) -> MetricsExportResult:
        try:
            responses = self.client.Export(
                self.generate_metrics_requests(export_records)
            )

            # Read response
            for _ in responses:
                pass

        except grpc.RpcError:
            return MetricsExportResult.FAILURE

        return MetricsExportResult.SUCCESS

    def shutdown(self) -> None:
        pass

    def generate_metrics_requests(
        self, metrics: Sequence[ExportRecord]
    ) -> metrics_service_pb2.ExportMetricsServiceRequest:
        collector_metrics = translate_to_collector(
            metrics, self.exporter_start_timestamp
        )
        service_request = metrics_service_pb2.ExportMetricsServiceRequest(
            node=self.node, metrics=collector_metrics
        )
        yield service_request


# pylint: disable=too-many-branches
def translate_to_collector(
    export_records: Sequence[ExportRecord],
    exporter_start_timestamp: Timestamp,
) -> Sequence[metrics_pb2.Metric]:
    collector_metrics = []
    for export_record in export_records:

        label_values = []
        label_keys = []
        for label_tuple in export_record.labels:
            label_keys.append(metrics_pb2.LabelKey(key=label_tuple[0]))
            label_values.append(
                metrics_pb2.LabelValue(
                    has_value=label_tuple[1] is not None,
                    value=str(label_tuple[1]),
                )
            )

        metric_descriptor = metrics_pb2.MetricDescriptor(
            name=export_record.instrument.name,
            description=export_record.instrument.description,
            unit=export_record.instrument.unit,
            type=get_collector_metric_type(export_record.instrument),
            label_keys=label_keys,
        )

        # If cumulative and stateful, explicitly set the start_timestamp to
        # exporter start time.
        if export_record.instrument.meter.processor.stateful:
            start_timestamp = exporter_start_timestamp
        else:
            start_timestamp = None

        timeseries = metrics_pb2.TimeSeries(
            label_values=label_values,
            points=[get_collector_point(export_record)],
            start_timestamp=start_timestamp,
        )
        collector_metrics.append(
            metrics_pb2.Metric(
                metric_descriptor=metric_descriptor,
                timeseries=[timeseries],
                resource=get_resource(export_record),
            )
        )
    return collector_metrics


# pylint: disable=no-else-return
def get_collector_metric_type(metric: Metric) -> metrics_pb2.MetricDescriptor:
    if isinstance(metric, Counter):
        if metric.value_type == int:
            return metrics_pb2.MetricDescriptor.CUMULATIVE_INT64
        elif metric.value_type == float:
            return metrics_pb2.MetricDescriptor.CUMULATIVE_DOUBLE
    return metrics_pb2.MetricDescriptor.UNSPECIFIED


def get_collector_point(export_record: ExportRecord) -> metrics_pb2.Point:
    # TODO: horrible hack to get original list of keys to then get the bound
    # instrument
    point = metrics_pb2.Point(
        timestamp=utils.proto_timestamp_from_time_ns(
            export_record.aggregator.last_update_timestamp
        )
    )
    if export_record.instrument.value_type == int:
        point.int64_value = export_record.aggregator.checkpoint
    elif export_record.instrument.value_type == float:
        point.double_value = export_record.aggregator.checkpoint
    else:
        raise TypeError(
            "Unsupported metric type: {}".format(
                export_record.instrument.value_type
            )
        )
    return point


def get_resource(export_record: ExportRecord) -> resource_pb2.Resource:
    resource_attributes = export_record.resource.attributes
    return resource_pb2.Resource(
        type=infer_oc_resource_type(resource_attributes),
        labels={k: str(v) for k, v in resource_attributes.items()},
    )


def infer_oc_resource_type(resource_attributes: Dict[str, str]) -> str:
    """Convert from OT resource labels to OC resource type"""
    for (
        ot_resource_key,
        oc_resource_type,
    ) in _OT_LABEL_PRESENCE_TO_RESOURCE_TYPE:
        if ot_resource_key in resource_attributes:
            return oc_resource_type
    return ""



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

import grpc
from google.protobuf.timestamp_pb2 import Timestamp
from opencensus.proto.metrics.v1 import metrics_pb2

from opentelemetry import metrics
from opentelemetry.exporter.opencensus import metrics_exporter
from opentelemetry.sdk.metrics import (
    Counter,
    MeterProvider,
    ValueRecorder,
    get_dict_as_key,
)
from opentelemetry.sdk.metrics.export import (
    ExportRecord,
    MetricsExportResult,
    aggregate,
)
from opentelemetry.sdk.resources import Resource


# pylint: disable=no-member
class TestCollectorMetricsExporter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # pylint: disable=protected-access
        cls._resource_labels = {
            "key_with_str_value": "some string",
            "key_with_int_val": 321,
            "key_with_true": True,
        }
        metrics.set_meter_provider(
            MeterProvider(resource=Resource(cls._resource_labels))
        )
        cls._meter = metrics.get_meter(__name__)
        cls._labels = {"environment": "staging", "number": 321}
        cls._key_labels = get_dict_as_key(cls._labels)

    def test_constructor(self):
        mock_get_node = mock.Mock()
        patch = mock.patch(
            "opentelemetry.exporter.opencensus.util.get_node",
            side_effect=mock_get_node,
        )
        service_name = "testServiceName"
        host_name = "testHostName"
        client = grpc.insecure_channel("")
        endpoint = "testEndpoint"
        with patch:
            exporter = metrics_exporter.OpenCensusMetricsExporter(
                service_name=service_name,
                host_name=host_name,
                endpoint=endpoint,
                client=client,
            )

        self.assertIs(exporter.client, client)
        self.assertEqual(exporter.endpoint, endpoint)
        mock_get_node.assert_called_with(service_name, host_name)

    def test_get_collector_metric_type(self):
        result = metrics_exporter.get_collector_metric_type(
            Counter("testName", "testDescription", "unit", int, None)
        )
        self.assertIs(result, metrics_pb2.MetricDescriptor.CUMULATIVE_INT64)
        result = metrics_exporter.get_collector_metric_type(
            Counter("testName", "testDescription", "unit", float, None)
        )
        self.assertIs(result, metrics_pb2.MetricDescriptor.CUMULATIVE_DOUBLE)
        result = metrics_exporter.get_collector_metric_type(
            ValueRecorder("testName", "testDescription", "unit", None, None)
        )
        self.assertIs(result, metrics_pb2.MetricDescriptor.UNSPECIFIED)

    def test_get_collector_point(self):
        aggregator = aggregate.SumAggregator()
        int_counter = self._meter.create_counter(
            "testName", "testDescription", "unit", int,
        )
        float_counter = self._meter.create_counter(
            "testName", "testDescription", "unit", float,
        )
        valuerecorder = self._meter.create_valuerecorder(
            "testName", "testDescription", "unit", float,
        )
        result = metrics_exporter.get_collector_point(
            ExportRecord(
                int_counter,
                self._key_labels,
                aggregator,
                metrics.get_meter_provider().resource,
            )
        )
        self.assertIsInstance(result, metrics_pb2.Point)
        self.assertIsInstance(result.timestamp, Timestamp)
        self.assertEqual(result.int64_value, 0)
        aggregator.update(123.5)
        aggregator.take_checkpoint()
        result = metrics_exporter.get_collector_point(
            ExportRecord(
                float_counter,
                self._key_labels,
                aggregator,
                metrics.get_meter_provider().resource,
            )
        )
        self.assertEqual(result.double_value, 123.5)
        self.assertRaises(
            TypeError,
            metrics_exporter.get_collector_point(
                ExportRecord(
                    valuerecorder,
                    self._key_labels,
                    aggregator,
                    metrics.get_meter_provider().resource,
                )
            ),
        )

    def test_export(self):
        mock_client = mock.MagicMock()
        mock_export = mock.MagicMock()
        mock_client.Export = mock_export
        host_name = "testHostName"
        collector_exporter = metrics_exporter.OpenCensusMetricsExporter(
            client=mock_client, host_name=host_name
        )
        test_metric = self._meter.create_counter(
            "testname", "testdesc", "unit", int, self._labels.keys(),
        )
        record = ExportRecord(
            test_metric,
            self._key_labels,
            aggregate.SumAggregator(),
            metrics.get_meter_provider().resource,
        )

        result = collector_exporter.export([record])
        self.assertIs(result, MetricsExportResult.SUCCESS)
        # pylint: disable=unsubscriptable-object
        export_arg = mock_export.call_args[0]
        service_request = next(export_arg[0])
        output_metrics = getattr(service_request, "metrics")
        output_node = getattr(service_request, "node")
        self.assertEqual(len(output_metrics), 1)
        self.assertIsNotNone(getattr(output_node, "library_info"))
        self.assertIsNotNone(getattr(output_node, "service_info"))
        output_identifier = getattr(output_node, "identifier")
        self.assertEqual(
            getattr(output_identifier, "host_name"), "testHostName"
        )

    def test_translate_to_collector(self):
        test_metric = self._meter.create_counter(
            "testname", "testdesc", "unit", int, self._labels.keys()
        )
        aggregator = aggregate.SumAggregator()
        aggregator.update(123)
        aggregator.take_checkpoint()
        record = ExportRecord(
            test_metric,
            self._key_labels,
            aggregator,
            metrics.get_meter_provider().resource,
        )
        start_timestamp = Timestamp()
        output_metrics = metrics_exporter.translate_to_collector(
            [record], start_timestamp,
        )
        self.assertEqual(len(output_metrics), 1)
        self.assertIsInstance(output_metrics[0], metrics_pb2.Metric)
        self.assertEqual(output_metrics[0].metric_descriptor.name, "testname")
        self.assertEqual(
            output_metrics[0].metric_descriptor.description, "testdesc"
        )
        self.assertEqual(output_metrics[0].metric_descriptor.unit, "unit")
        self.assertEqual(
            output_metrics[0].metric_descriptor.type,
            metrics_pb2.MetricDescriptor.CUMULATIVE_INT64,
        )
        self.assertEqual(
            len(output_metrics[0].metric_descriptor.label_keys), 2
        )
        self.assertEqual(
            output_metrics[0].metric_descriptor.label_keys[0].key,
            "environment",
        )
        self.assertEqual(
            output_metrics[0].metric_descriptor.label_keys[1].key, "number",
        )

        self.assertIsNotNone(output_metrics[0].resource)
        self.assertEqual(
            output_metrics[0].resource.type, "",
        )
        self.assertEqual(
            output_metrics[0].resource.labels["key_with_str_value"],
            self._resource_labels["key_with_str_value"],
        )
        self.assertIsInstance(
            output_metrics[0].resource.labels["key_with_int_val"], str,
        )
        self.assertEqual(
            output_metrics[0].resource.labels["key_with_int_val"],
            str(self._resource_labels["key_with_int_val"]),
        )
        self.assertIsInstance(
            output_metrics[0].resource.labels["key_with_true"], str,
        )
        self.assertEqual(
            output_metrics[0].resource.labels["key_with_true"],
            str(self._resource_labels["key_with_true"]),
        )

        self.assertEqual(len(output_metrics[0].timeseries), 1)
        self.assertEqual(len(output_metrics[0].timeseries[0].label_values), 2)
        self.assertEqual(
            output_metrics[0].timeseries[0].start_timestamp, start_timestamp
        )
        self.assertEqual(
            output_metrics[0].timeseries[0].label_values[0].has_value, True
        )
        self.assertEqual(
            output_metrics[0].timeseries[0].label_values[0].value, "staging"
        )
        self.assertEqual(len(output_metrics[0].timeseries[0].points), 1)
        self.assertEqual(
            output_metrics[0].timeseries[0].points[0].timestamp.seconds,
            record.aggregator.last_update_timestamp // 1000000000,
        )
        self.assertEqual(
            output_metrics[0].timeseries[0].points[0].timestamp.nanos,
            record.aggregator.last_update_timestamp % 1000000000,
        )
        self.assertEqual(
            output_metrics[0].timeseries[0].points[0].int64_value, 123
        )

    def test_infer_ot_resource_type(self):
        # empty resource
        self.assertEqual(metrics_exporter.infer_oc_resource_type({}), "")

        # container
        self.assertEqual(
            metrics_exporter.infer_oc_resource_type(
                {
                    "k8s.cluster.name": "cluster1",
                    "k8s.pod.name": "pod1",
                    "k8s.namespace.name": "namespace1",
                    "container.name": "container-name1",
                    "cloud.account.id": "proj1",
                    "cloud.zone": "zone1",
                }
            ),
            "container",
        )

        # k8s pod
        self.assertEqual(
            metrics_exporter.infer_oc_resource_type(
                {
                    "k8s.cluster.name": "cluster1",
                    "k8s.pod.name": "pod1",
                    "k8s.namespace.name": "namespace1",
                    "cloud.zone": "zone1",
                }
            ),
            "k8s",
        )

        # host
        self.assertEqual(
            metrics_exporter.infer_oc_resource_type(
                {
                    "k8s.cluster.name": "cluster1",
                    "cloud.zone": "zone1",
                    "host.name": "node1",
                }
            ),
            "host",
        )

        # cloud
        self.assertEqual(
            metrics_exporter.infer_oc_resource_type(
                {
                    "cloud.provider": "gcp",
                    "host.id": "inst1",
                    "cloud.zone": "zone1",
                }
            ),
            "cloud",
        )
